#include "AneuMeshLoader.h"
#include "Exceptions.h"

#include <iostream>
#include <fstream>
#include <vector>

	void AneuMeshLoader::LoadMesh(const std::string& filename) {
		 std::ifstream input(filename);
		 if (!input) {
			 throw NoFileFoundException(filename);
		 }
		 else std::cout << "File \"" << filename << "\" was succesfully opened" << std::endl;

		 int dimension, amount;

		 input >> amount >> dimension;
		 this->nodes.reserve(amount);
		 for (int i = 1; i <= amount; ++i)
		 {
			 Node temp_node;
			 temp_node.cords.reserve(dimension);
			 temp_node.nodeID = i;
			 temp_node.vertex = true;
			 double coord;
			 for (int j = 0; j < dimension; j++)
			 {
				 input >> coord;
				 temp_node.cords.push_back(coord);
			 }
			 this->nodes.push_back(temp_node);
		 }
	 
		 input >> amount >> dimension;
		 this->elements.reserve(amount);
		 for (int i = 1; i <= amount; ++i) 
		 {
			 FiniteElement temp_element;
			 temp_element.elementID = i;
			 input >> temp_element.materialID;
			 temp_element.nodesID.reserve(dimension);
			 int temp_id;
			 for (int j = 0; j < dimension; ++j) {
				 input >> temp_id;
				 temp_element.nodesID.push_back(temp_id);
			 }
			 this->elements.push_back(std::move(temp_element));
		 }
	 
		 input >> amount >> dimension;
		 this->boundaryFEs.reserve(amount);
		 for (int i = 1; i <= amount; ++i) {
			 BoundaryFiniteElement temp_surface;
			 temp_surface.surfaceID = i;
			 input >> temp_surface.boundaryID;
			 temp_surface.nodesID.reserve(dimension);
			 int temp_id;
			 for (int j = 0; j < dimension; ++j) {
				 input >> temp_id;
				 temp_surface.nodesID.push_back(temp_id);
			 }
			 this->boundaryFEs.push_back(std::move(temp_surface));
		 }

		 input.close();


		 std::cout << "All data was loaded, closing the file..." << std::endl;
	}