﻿#include <algorithm>
#include <stdlib.h>
#include <unordered_set>
#include <iterator>
#include <map>
#include "MeshLoader.h"
#include <fstream>
#include <ranges>

std::vector<Node> MeshLoader::getNodes() 
{
	return nodes;
}

std::vector<FiniteElement> MeshLoader::getFiniteElements() 
{
	return elements;
}

std::vector<BoundaryFiniteElement> MeshLoader::getBoundaryFiniteElements() 
{
	return boundaryFEs;
}


//!!! Контенеры-результаты должны содержать только ID узлов или КЭ


std::vector<FiniteElement> MeshLoader::getElementsByID(const int id1, const int id2, const int id3) //возможность нескольких узлов - контейнер
{
	std::vector<FiniteElement> result;
	auto current_iter = elements.begin();
	while (current_iter != elements.end())
	{
		current_iter = std::find_if(current_iter, elements.end(), [id1, id2, id3](const FiniteElement& current_element) {
			return std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id1) != current_element.nodesID.end()
				&& std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id2) != current_element.nodesID.end()
				&& std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id3) != current_element.nodesID.end(); });
		
		if (current_iter != elements.end())
		{
			result.push_back(*current_iter);
			current_iter++;
		}
	}
	return result;//O(n)
}

std::vector<FiniteElement> MeshLoader::getElementsByEdge(const int id1, const int id2) //возможность нескольких КЭ 
{
	/*std::vector<FiniteElement> result;
	std::copy_if(elements.begin(), elements.end(), std::back_inserter(result), [id1, id2](const FiniteElement& current_element) {
		return std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id1) != current_element.nodesID.end()
			&& std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id2) != current_element.nodesID.end(); });
	return result;*/
	std::vector<FiniteElement> result;
	auto   last = boundaryFEs.begin();
	while (last != boundaryFEs.end())
	{
		last = find_if(last, boundaryFEs.end(), [id1, id2](const FiniteElement& elem)
			{
				return (std::find(elem.nodesID.begin(), elem.nodesID.end(), id1) != elem.nodesID.end()
					&& std::find(elem.nodesID.begin(), elem.nodesID.end(), id2) != elem.nodesID.end());
			});

		if (last != boundaryFEs.end())
			result.push_back(*last++);
	}

	return result;

}

std::vector<FiniteElement> MeshLoader::getElementsByOneID(const int id1) 
{
	std::vector<FiniteElement> result;
	std::copy_if(elements.begin(), elements.end(), std::back_inserter(result), [id1](const FiniteElement& current_element) {
		return std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id1) != current_element.nodesID.end(); });
	return result;
	//O(n)
}

std::vector<Node> MeshLoader::getNodesByBoundary(int boundaryID) 
{
	std::set<Node> result;
	auto   last = boundaryFEs.begin();
	while (last != boundaryFEs.end()) {
		last = std::find_if(last, boundaryFEs.end(), [boundaryID](const BoundaryFiniteElement& elem) {
			return elem.boundaryID == boundaryID; });
		if (last != boundaryFEs.end()) {
			for (const auto& it : last->nodesID) {
				auto node_by_ID = std::find_if(nodes.begin(), nodes.end(), [it](const Node& p_Node) { return p_Node.nodeID == it; });
				result.insert(*node_by_ID);
			}
			++last;
		}
	}
	return std::vector<Node>(result.begin(), result.end());
}

std::vector<FiniteElement> MeshLoader::getElementsByBoundary(int surfaceID) 
{
	
	std::vector<FiniteElement> result;//vector?
	std::copy_if(boundaryFEs.begin(), boundaryFEs.end(), std::back_inserter(result), [surfaceID](const BoundaryFiniteElement& current_element) {
		return surfaceID == current_element.surfaceID;
		});
	return result;
	/*
	std::set<FiniteElement> result;
	auto last = boundaryFEs.begin();
	while (last != boundaryFEs.end()) {
		last = std::find_if(last, boundaryFEs.end(), [surfaceID](const BoundaryFiniteElement& elem) {
			return elem.boundaryID == surfaceID; });
		if (last != boundaryFEs.end()) {
			auto element_by_ID = std::find_if(elements.begin(), elements.end(), [last](FiniteElement& p_FiniteElement) {
				return p_FiniteElement.elementID == last->elementID; });
			result.insert(*element_by_ID);
			++last;
		}
	}
	return std::vector<FiniteElement>(result.begin(), result.end());
	*/
}

std::vector<BoundaryFiniteElement> MeshLoader::getsboundaryFEsByBoundary(int boundaryID)//что поменять?
{
	std::vector<BoundaryFiniteElement> result;
	auto last = boundaryFEs.begin();
	while (last != boundaryFEs.end()) {
		last = std::find_if(last, boundaryFEs.end(), [boundaryID](const BoundaryFiniteElement& elem) {
			return elem.boundaryID == boundaryID; });
		if (last != boundaryFEs.end()) {
			result.push_back(*last);
			++last;
		}
	}
	return result;

}

void MeshLoader::insertCenterNode() {
	
	std::unordered_set<Edge> edges;

	for (auto& it : elements) {
		std::vector<int> current_nodes(it.nodesID);
		for (auto i = current_nodes.begin(); i < current_nodes.end() - 1; ++i)
			for (auto j = i + 1; j < current_nodes.end(); ++j) {
				Edge current_edge(*i, *j, 0);

				auto founded_edge = edges.find(current_edge);

				if (founded_edge == edges.cend()) {
					Node left_node = nodes[*i - 1];
					Node right_node = nodes[*j - 1];
					int center_id_ = nodes.rbegin()->nodeID + 1;
					Node center_node(center_id_, (left_node.cords[0] + right_node.cords[0]) / 2,
						(left_node.cords[1] + right_node.cords[1]) / 2, (left_node.cords[2] + right_node.cords[2]) / 2, false);

					current_edge.updateCenter(center_id_);
					nodes.push_back(center_node);
					edges.insert(current_edge);
					it.nodesID.push_back(center_id_);
				}
				else {
					it.nodesID.push_back(founded_edge->center_id);
				}
			}
		it.nodesID.shrink_to_fit();
	}



	for (auto& it : boundaryFEs) {
		std::vector<int> current_nodes(it.nodesID);
		for (auto i = current_nodes.begin(); i < current_nodes.end() - 1; ++i)
			for (auto j = i + 1; j < current_nodes.end(); ++j) {
				Edge current_edge(*i, *j, 0);

				auto founded_edge = edges.find(current_edge);

				if (founded_edge != edges.cend())
					it.nodesID.push_back(founded_edge->center_id);
			}
		it.nodesID.shrink_to_fit();
	}

	boundaryFEs.shrink_to_fit();
	nodes.shrink_to_fit();

}


std::vector<std::set<int>> MeshLoader::getNodeNeighbours()
{
	std::vector<std::set<int>> result(getNodes().size());

	for (auto& i : elements) {
		for (int j : i.nodesID) {
			result[j - 1].insert(i.nodesID.begin(), i.nodesID.end());
			for (std::set<int>::iterator iter = result[j - 1].begin(); iter != result[j - 1].end();)
			{
				if (*iter == j)
				{
					iter = result[j - 1].erase(iter);
				}
				else
				{
					++iter;
				}
			}
		}
	}
	return result;

}


void MeshLoader::Print_Data() 
{
	std::cout << getNodes() << std::endl
		<< getFiniteElements() << std::endl
		<< getBoundaryFiniteElements() << std::endl;
}


const std::vector<double>& MeshLoader::findNodeRanges(int id)
{
	return (std::ranges::find_if(nodes.begin(), nodes.end(), [id](const Node& elem)
		{
			return elem.nodeID == id;
		}))->cords;
}


std::vector<Node>::iterator MeshLoader::getNodeByID(int id)
{
	Node tmp;
	tmp.nodeID = id;
	return find(nodes.begin(), nodes.end(), tmp);
}



std::vector<BoundaryFiniteElement> MeshLoader::getTrianglesAndNodes(int surfaceID)
{
	std::vector<BoundaryFiniteElement> result;//vector?
	std::copy_if(boundaryFEs.begin(), boundaryFEs.end(), std::back_inserter(result), 
	[surfaceID](const BoundaryFiniteElement& current_element) {
		return surfaceID == current_element.surfaceID;
		});
	auto square = [](Node A, Node B, Node C)
	{
		
		double a = sqrt((C.cords[0] - B.cords[0]) * (C.cords[0] - B.cords[0]) + (C.cords[1] - B.cords[1]) * (C.cords[1] - B.cords[1]) + (C.cords[2] - B.cords[2]) * (C.cords[2] - B.cords[2]));
		double b = sqrt((C.cords[0] - A.cords[0]) * (C.cords[0] - A.cords[0]) + (C.cords[1] - A.cords[1]) * (C.cords[1] - A.cords[1]) + (C.cords[2] - A.cords[2]) * (C.cords[2] - A.cords[2]));
		double c = sqrt((B.cords[0] - A.cords[0]) * (B.cords[0] - A.cords[0]) + (B.cords[1] - A.cords[1]) * (B.cords[1] - A.cords[1]) + (B.cords[2] - A.cords[2]) * (B.cords[2] - A.cords[2]));
		double p = 1 / 2 * (a + b + c);
		return sqrt(p * (p - a) * (p - b) * (p - c));
	};
	
	std::ranges::sort(result, [&](const BoundaryFiniteElement& a, const BoundaryFiniteElement& b)
		{
			Node X1 = *getNodeByID(a.nodesID[0]);
			Node Y1 = *getNodeByID(a.nodesID[1]);
			Node Z1 = *getNodeByID(a.nodesID[2]);

			Node X2 = *getNodeByID(b.nodesID[0]);
			Node Y2 = *getNodeByID(b.nodesID[1]);
			Node Z2 = *getNodeByID(b.nodesID[2]);

			return square(X1,Y1,Z1) > square(X2,Y2,Z2); 
		});
	std::ofstream triangle("triangle.txt", std::ofstream::out | std::ofstream::trunc);

	std::ostream_iterator<BoundaryFiniteElement> os(triangle, " ");
	int counter = 0;
	std::ranges::copy_if(result, std::ostream_iterator<BoundaryFiniteElement>(triangle, "\n"),
		[&counter](const auto& x) 
		{ 
		counter++;
		return ((counter - 1) % 20) >= 1; 
		});
	triangle.close();

	return result;//O(n)
}
